"""
API for ppv_cube
"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013-2017, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt_astro_analysis.ppv_cube.ppv_cube import \
    PPVCube
